/*
 * Simon Game
 * Copyright (C) 2006 Vin Helminen
 * $Id: simon.c,v 1.1.1.1 2006-12-11 04:27:11 vah Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "gp2x.h"

#define SCREEN_WIDTH  320
#define SCREEN_HEIGHT 240

#define DATA_DIR "data"

static const char* IMAGE_FILENAMES[] = {
  "background.png",
  "a_on.png", 
  "b_on.png", 
  "x_on.png", 
  "y_on.png", 
  "numbers.png",
};

enum {
  BACKGROUND = 0,
  A_ON,
  B_ON,
  X_ON,
  Y_ON,
  NUMBERS,
  IMAGE_COUNT,
};

static const char *SPLASHSCREEN = "splash.png";

/* some semirandom coordinates so that the images line up nicely */
#define A_X 100
#define A_Y 50
#define B_X 218
#define B_Y 48
#define X_X 128
#define X_Y 139
#define Y_X 129
#define Y_Y 19
#define NUMBERS_X 12
#define NUMBERS_Y 209

enum { A = 0, B, X, Y };

#define NUMBER_WIDTH   16
#define NUMBER_HEIGHT  20
#define NUMBER_SPACING 18

#define MAX_SEQUENCE_LENGTH 99

static inline void randomize(void);

static inline SDL_Surface** loadImages(void);

static inline void freeImages(SDL_Surface **images);

static void blitImage(SDL_Surface *image, SDL_Surface *screen, int x, int y); 

static inline void clear(SDL_Surface *screen);

static inline void paintNumbers(SDL_Surface *screen, int x, int y, const char *s, SDL_Surface *numbers);

int main(void)
{
  GP2XInit();
  GP2XCpuFrequency(200);

  chdir(DATA_DIR); /* change current directory so we can load the data */

  if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK) < 0) {
    fprintf(stderr, "SDL_Init: %s\n", SDL_GetError());
    return EXIT_FAILURE;
  }
  
  SDL_Surface *screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16, SDL_SWSURFACE);
  if(!screen) {
    fprintf(stderr, "SDL_SetVideoMode: %s\n", SDL_GetError());
    SDL_Quit();
    return EXIT_FAILURE;
  }
  SDL_ShowCursor(SDL_DISABLE);

  SDL_Joystick *joystick = SDL_JoystickOpen(0);
  if(!joystick) {
    fprintf(stderr, "SDL_JoystickOpen: %s\n", SDL_GetError());
    SDL_FreeSurface(screen);
    SDL_Quit();
    return EXIT_FAILURE;
  }

  SDL_Surface **images = NULL;

 reset:
  {
    clear(screen);
    SDL_Surface *splash = IMG_Load(SPLASHSCREEN);
    if(splash) {
      int w = splash->w, h = splash->h;
      int x = (SCREEN_WIDTH-w)>>1, y = (SCREEN_HEIGHT-h)>>1;
      blitImage(splash, screen, x, y);
      SDL_Flip(screen);
      SDL_FreeSurface(splash);
    }
  }
  if(!images) {
    images = loadImages();
    if(!images) {
      SDL_FreeSurface(screen);
      SDL_Quit();
      return EXIT_FAILURE;
    }
    GP2XCpuFrequency(50);
  }

  int lights[4] = {0, 0, 0, 0};
  enum {SPLASH, INTRO, COMPUTER, PLAYER, PL2CPU} mode = SPLASH;
  int sequence[MAX_SEQUENCE_LENGTH];
  int pos = 0, length = 0;
  int pressed = 0, lastpressed = 0;

  while(1) {
    SDL_Event event;
    while(SDL_PollEvent(&event)) {
      switch(event.type) {

      case SDL_QUIT:
	goto exit;

      case SDL_JOYBUTTONDOWN:
	if(event.jbutton.button == GP2X_BUTTON_SELECT)
	  goto exit;

	switch(mode) {
	case SPLASH:
	  mode = INTRO;
	  break;
	case INTRO:
	  if(event.jbutton.button == GP2X_BUTTON_START) {
	    mode = COMPUTER;
	    pos = length = 0;
	    lights[0] = lights[1] = lights[2] = lights[3] = 1;
	  }
	  break;
	case COMPUTER:
	  if(event.jbutton.button == GP2X_BUTTON_START)
	    goto reset;
	  break;
	case PLAYER:
	  switch(event.jbutton.button) {
	  case GP2X_BUTTON_START:
	    goto reset;
	    
	  case GP2X_BUTTON_A:
	    lights[A] = 1;
	    pressed |= 1<<A;
	    break;
	  case GP2X_BUTTON_B:
	    lights[B] = 1;
	    pressed |= 1<<B;
	    break;
	  case GP2X_BUTTON_X:
	    lights[X] = 1;
	    pressed |= 1<<X;
	    break;
	  case GP2X_BUTTON_Y:
	    lights[Y] = 1;
	    pressed |= 1<<Y;
	    break;
	  default:
	    break;
	  } /* switch button */
	  break;
	default:
	  break;
	} /* switch mode */
	break;

      case SDL_JOYBUTTONUP:
	if(mode == PLAYER) {
	  switch(event.jbutton.button) {
	  case GP2X_BUTTON_A:
	    lights[A] = 0;
	    pressed &= ~(1<<A);
	    break;
	  case GP2X_BUTTON_B:
	    lights[B] = 0;
	    pressed &= ~(1<<B);
	    break;
	  case GP2X_BUTTON_X:
	    lights[X] = 0;
	    pressed &= ~(1<<X);
	    break;
	  case GP2X_BUTTON_Y:
	    lights[Y] = 0;
	    pressed &= ~(1<<Y);
	    break;
	  default:
	    break;
	  } /* switch button */
	}
	break;
      default:
	break;
      } /* switch event type */
      
    } /* if event */

    if(mode != SPLASH) {
      blitImage(images[BACKGROUND], screen, 0, 0);
      if(lights[A]) blitImage(images[A_ON], screen, A_X, A_Y);
      if(lights[B]) blitImage(images[B_ON], screen, B_X, B_Y);
      if(lights[X]) blitImage(images[X_ON], screen, X_X, X_Y);
      if(lights[Y]) blitImage(images[Y_ON], screen, Y_X, Y_Y);      
      {
	char s[6];
	sprintf(s, "%2d/%2d", pos, length);
	paintNumbers(screen, NUMBERS_X, NUMBERS_Y, s, images[NUMBERS]);
      }
      GP2XWaitVBlank();
      SDL_Flip(screen);
    }

    if(mode != PLAYER) {
      SDL_Delay(250-(length<<1));
    } 

    switch(mode) {
    case SPLASH:
      break;
    case INTRO:
      {
	int i = rand()%4;
	lights[i] = !lights[i];
      }
      break;
    case COMPUTER:
      {
	int wait = lights[0] || lights[1] || lights[2] || lights[3];
	if(wait) {
	  lights[0] = lights[1] = lights[2] = lights[3] = 0;
	} else {
	  if(pos == 0) length++;
	  if(pos < length-1) {
	    lights[sequence[pos]] = 1;
	    pos++;
	  } else if(pos == length-1) {
	    int i = (rand()>>8)%4;
	    lights[i] = 1;
	    sequence[pos] = i;
	    pos++;
	  } else {
	    mode = PLAYER;
	    pressed = lastpressed = 0;
	    pos = 0;
	  }
	} /* wait */
      }
      break;
    case PLAYER:
      {
	if(pressed && pressed != lastpressed) {
	  int sum = lights[0] + lights[1] + lights[2] + lights[3];
	  if(sum == 1) {
	    int i;
	    for(i = 0; pressed >> i; ++i);
	    --i;
	    if(i != sequence[pos]) sum = 0; // player pushed a wrong button
	    else pos++;
	    if(pos >= length)
	      mode = PL2CPU;
	  }
	  if(sum != 1) {
	    // game over 
	    mode = INTRO;
	    lights[0] = lights[1] = lights[2] = lights[3] = 1;
	  }
	}
	lastpressed = pressed;
      }
      break;
    case PL2CPU:
      lights[0] = lights[1] = lights[2] = lights[3] = 0;
      if(pos == MAX_SEQUENCE_LENGTH) {
	// player has won
	mode = INTRO;
      } else {
	pos = 0;
	mode = COMPUTER;
      }
      break;
    }

  } /* while 1 */
 exit:
  freeImages(images);
  SDL_FreeSurface(screen);
  SDL_Quit();
  return EXIT_SUCCESS;
}

static inline void randomize(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srand((tv.tv_usec) | (tv.tv_sec));
}


static inline SDL_Surface** loadImages()
{
  SDL_Surface **images = calloc(IMAGE_COUNT, sizeof(SDL_Surface*));
  if(!images) {
    fprintf(stderr, "calloc failed.\n");
    return NULL;
  }
  for(int i = 0; i < IMAGE_COUNT; ++i) {
    SDL_Surface *img = NULL, *tmp = IMG_Load(IMAGE_FILENAMES[i]);
    if(tmp) {
      if(i != BACKGROUND) {
	Uint32 colorkey = SDL_MapRGB(tmp->format, 0, 0, 0); /* black is transparent */
	SDL_SetColorKey(tmp, SDL_RLEACCEL|SDL_SRCCOLORKEY, colorkey);
      }
      img = SDL_DisplayFormat(tmp);
      SDL_FreeSurface(tmp);
    }
    if(!img) {
      fprintf(stderr, "loadImages: %s\n", SDL_GetError());
      while(--i >= 0) SDL_FreeSurface(images[i]);
      free(images);
      return NULL;
    }   
    images[i] = img;
  }
  return images;
}

static inline void freeImages(SDL_Surface **images)
{
  for(int i = 0; i < IMAGE_COUNT; ++i)
    SDL_FreeSurface(images[i]);
  free(images);
}

static void blitImage(SDL_Surface *image, SDL_Surface *screen, int x, int y)
{
  SDL_Rect src, dst;
  src.x = 0; src.y = 0;
  src.w = image->w; src.h = image->h;
  dst.x = x; dst.y = y;
  SDL_BlitSurface(image, &src, screen, &dst);
}

static inline void clear(SDL_Surface *screen)
{
  SDL_Rect r;
  r.x = r.y =0;
  r.w = screen->w;
  r.h = screen->h;
  SDL_FillRect(screen, &r, SDL_MapRGB(screen->format, 0, 0, 0));
}

static inline void paintNumbers(SDL_Surface *screen, int x, int y, const char *s, SDL_Surface *numbers)
{
  do {
    int c = *s, n = -1;
    if(c == '\0') break;
    if(c >= '0' && c <= '9') n = c - '0';
    else if(c == '/') n = 10;
    if(n >= 0) {
      SDL_Rect src, dst;
      src.x = NUMBER_WIDTH*n;
      src.y = 0;
      src.w = NUMBER_WIDTH;
      src.h = NUMBER_HEIGHT;
      dst.x = x;
      dst.y = y;
      SDL_BlitSurface(numbers, &src, screen, &dst);
    }
    x += NUMBER_SPACING;
    ++s;
  } while(1);
}
